#!/usr/local/bin/python

import _kinfxdb
import operator

error=_kinfxdb.error
dbi=_kinfxdb.dbi

def connect(*arg):
	return apply(InfxConnectionObject, arg)
        
class InfxConnectionObject:
	def __init__(self, dbname, user=None, passwd=None):
		self.conn, isonline=_kinfxdb.attach_db(dbname, user, passwd)
                if isonline: self.server='ON'
                else: self.server='SE'

        def __del__(self):
        	if hasattr(self, "conn"):
	        	_kinfxdb.detach_db(self.conn)

        def begin(self, param=None):
        	"""start transaction"""
	        _kinfxdb.begin(self.conn)

        def commit(self):
        	"""commit transaction"""
                _kinfxdb.commit(self.conn)

        def rollback(self):
        	"""commit transaction"""
                _kinfxdb.rollback(self.conn)

        def execute_immediate(self, sql):
        	"""fast execution"""
		_kinfxdb.execute_immediate(self.conn, sql)

        def cursor(self):
        	"""get cursor object"""
		return InfxCursor(self.conn)

        def close(self):
        	"""closes connection to server"""
        	if hasattr(self, "conn"):
	        	_kinfxdb.detach_db(self.conn)
                        del self.conn

class InfxCursor:
	def __init__(self, conn):
        	self.conn=conn
                self.cur=_kinfxdb.cursor(self.conn)
                self.description=()

        def __del__(self):
        	if hasattr(self, "cur"):
			_kinfxdb.free_cursor(self.cur)

        def execute(self, sql, params=None):
        	res=(-1, ())
        	if params==None:
        		res=_kinfxdb.execute(self.cur, sql)
                else:
                	if type(params)==type([]):
                        	for param in params:
                                	res=_kinfxdb.execute(self.cur, sql, param)
                        else:
	                	res=_kinfxdb.execute(self.cur, sql, params)
                self.description=res[1]
# if is description then replaceing column name 
# if name equal 'expression' then rename 'expression_%count' because of zope //papa
		if self.description is not None:
		    ex_num=0
		    buf2=[]
		    for seq in range(len(self.description)):
			buf1=[]
			for b1item in self.description[seq]:
			    buf1.append(b1item)
			if buf1[0]=='(expression)':
			    buf1[0]='(expression)_%i' % (ex_num)
			    ex_num+=1
			buf2.append(tuple(buf1))
		    self.description=tuple(buf2)
# end of modify //papa
                return res[0]

        def fetch(self):
        	return _kinfxdb.fetch(self.cur)

        def fetchone(self):
        	return self.fetch()

        def fetchall(self):
        	res=None
                all=[]
                while 1:
                	res=_kinfxdb.fetch(self.cur)
                        if res==None:
                        	break
                        all.append(res)
        	return all

        def fetchmany(self, count):
        	res=None
                all=[]
                i=0
                while i<count:
                	res=_kinfxdb.fetch(self.cur)
                        if res==None:
                        	break
                        all.append(res)
        	return all

        def close(self):
        	return _kinfxdb.close(self.cur)


def infxtest():
	execfile('script/test.py')

if __name__=='__main__':
	infxtest()
