/*
 dbimodule.c
 $Date: 1998/09/07 13:40:06 $
 
 This is the general interface to COPPERMAN compliant database
 interface.
 
 In particular, types and type numbers are defined
 Extended by Alexander Kuznetsov, 1998
 */

#include <time.h>

#include "Python.h"
#include "intobject.h"

#include "dbi.h"

/*
#define VERBOSE
  */
typedef struct
{
    PyObject_HEAD
        PyObject *objectOf;
} DbiContainer;


PyObject *dbiValue(PyObject *o)
{
    return ((DbiContainer *) o)->objectOf;
}

static PyObject *makeDbiTypeTP(PyObject *args, PyTypeObject *ty)
{
    DbiContainer *dbt = 0;
    PyObject *ob;
    
    if (PyArg_ParseTuple(args, "O", &ob)) {
        dbt = PyObject_NEW(DbiContainer, ty);
        dbt->objectOf = ob;
        Py_INCREF(ob);
    }
    return (PyObject*)dbt;
}

static void dbiDealloc(PyObject *self)
{
    Py_DECREF(((DbiContainer *) self)->objectOf);
    PyMem_DEL(self);
}

static PyMethodDef noMethods[] = {
    0, 0
};

static PyObject *dbiGetAttr(PyObject *self,
                            char *name)
{
    if (!strcmp(name, "value")) {
        PyObject *val = dbiValue(self);
        Py_INCREF(val);
        return val;
    }
    return Py_FindMethod(noMethods, self, name);
}

static PyObject *dateStr(PyObject *o)
{
    char *t;
    long l = PyInt_AsLong(dbiValue(o));
    t=ctime((const time_t *)&l);
    return PyString_FromStringAndSize(t, strlen(t)-1); /* less \n */
}

static PyObject *dateRepr(PyObject *o)
{
    char buf[80];
    char *t;
    long l = PyInt_AsLong(dbiValue(o));
    t=ctime((const time_t *)&l);
    t[strlen(t)-1]=0;
    sprintf(buf, "<dbiDate '%s'>", t);
    return PyString_FromString(buf); /* less \n */
}

#define delg(a) dbiValue(a)->ob_type->tp_as_number

static PyObject* dt_nb_add(PyObject* a, PyObject* b)
{ return delg(a)->nb_add(dbiValue(a),b); }
static PyObject* dt_nb_subtract(PyObject* a, PyObject* b)
{ return delg(a)->nb_subtract(dbiValue(a),b); }
static PyObject* dt_nb_int(PyObject* a)
{ return delg(a)->nb_int(dbiValue(a)); }
static PyObject* dt_nb_long(PyObject* a)
{ return delg(a)->nb_long(dbiValue(a)); }
static PyObject* dt_nb_float(PyObject* a)
{ return delg(a)->nb_float(dbiValue(a)); }
static PyObject* dt_nb_oct(PyObject* a)
{ return delg(a)->nb_oct(dbiValue(a)); }
static PyObject* dt_nb_hex(PyObject* a)
{ return delg(a)->nb_hex(dbiValue(a)); }
static int dt_cmp(PyObject *a, PyObject *b)
{ return dbiValue(a)->ob_type->tp_compare(a,b); }
static int dt_nb_coerce(PyObject **, PyObject **);

PyNumberMethods dt_as_number = {
    dt_nb_add ,   /* nb_add */
    dt_nb_subtract ,   /* nb_subtract */
    0,   /* nb_multiply */
    0,   /* nb_divide */
    0,   /* nb_remainder */
    0,   /* nb_divmod */
    0,   /* nb_power */
    0,   /* nb_negative */
    0,   /* nb_positive */
    0,   /* nb_absolute */
    0,   /* nb_nonzero */
    0,   /* nb_invert */
    0,   /* nb_lshift */
    0,   /* nb_rshift */
    0,   /* nb_and */
    0,   /* nb_xor */
    0,   /* nb_or */
    dt_nb_coerce,   /* nb_coerce */
    dt_nb_int ,   /* nb_int */
    dt_nb_long ,   /* nb_long */
    dt_nb_float ,   /* nb_float */
    dt_nb_oct ,   /* nb_oct */
    dt_nb_hex    /* nb_hex */
} ;

static PyTypeObject DbiDate_Type =
{
    PyObject_HEAD_INIT (&PyType_Type)
        0,			/*ob_size */
        "DbiDate",		/*tp_name */
        sizeof(DbiContainer),	/*tp_basicsize */
        0,			/*tp_itemsize */
        dbiDealloc,		/*tp_dealloc */
        0,			/*tp_print */
        dbiGetAttr,		/*tp_getattr */
        0,                    /*tp_setattr */
        dt_cmp,               /*tp_compare */
        dateRepr,              /*tp_repr */
        &dt_as_number,        /**tp_as_number */
        0,                    /**tp_as_sequence */
        0,                    /**tp_as_mapping */
        0,                    /*tp_hash */
        0,                    /*tp_call */
        dateStr               /*tp_str */
        
} ;
static int dt_nb_coerce(PyObject **pv, PyObject **pw)
{
    if (PyNumber_Check(*pw)) {
        DbiContainer *dbt = PyObject_NEW(DbiContainer, &DbiDate_Type);
        dbt->objectOf = PyNumber_Int(*pw);
        Py_INCREF(*pv);
        *pw = (PyObject *)dbt;
        return 0;
    }
    return 1; /* Can't do it */
}

static PyTypeObject DbiRaw_Type =
{
    PyObject_HEAD_INIT (&PyType_Type)
        0,			/*ob_size */
        "DbiRaw",		/*tp_name */
        sizeof(DbiContainer),	/*tp_basicsize */
        0,			/*tp_itemsize */
        dbiDealloc,   	/*tp_dealloc */
        0,			/*tp_print */
        dbiGetAttr,		/*tp_getattr */
        0,                    /*tp_setattr */
        0,                    /*tp_compare */
        0,                    /*tp_repr */
        0,                    /**tp_as_number */
        0,                    /**tp_as_sequence */
        0,                    /**tp_as_mapping */
        0,                    /*tp_hash */
        0,                    /*tp_call */
        dbiValue              /*tp_str */
} ;

static PyTypeObject DbiRowId_Type =
{
    PyObject_HEAD_INIT (&PyType_Type)
        0,			/*ob_size */
        "DbiRowId",		/*tp_name */
        sizeof(DbiContainer),	/*tp_basicsize */
        0,			/*tp_itemsize */
        dbiDealloc,   	/*tp_dealloc */
        0,			/*tp_print */
        dbiGetAttr		/*tp_getattr */
} ;

static PyObject *makeDate(PyObject *self, PyObject *args)
{
    return makeDbiTypeTP(args, &DbiDate_Type);
}

static PyObject *makeRaw(PyObject *self, PyObject *args)
{
    return makeDbiTypeTP(args, &DbiRaw_Type);
}

static PyObject *makeRowId(PyObject *self, PyObject *args)
{
    return makeDbiTypeTP(args, &DbiRowId_Type);
}

static PyObject *makeDbiType(PyObject *o, PyTypeObject *ty)
{
    DbiContainer *dbt = PyObject_NEW(DbiContainer, ty);
    dbt->objectOf = o;
    return (PyObject*) dbt;
}

PyObject *dbiMakeDate(PyObject *contents)
{
    return makeDbiType(contents, &DbiDate_Type);
}

PyObject *dbiMakeRaw(PyObject *contents)
{
    return makeDbiType(contents, &DbiRaw_Type);
}

PyObject *dbiMakeRowId(PyObject *contents)
{
    return makeDbiType(contents, &DbiRowId_Type);
}



static PyMethodDef dbiMethods[] = {
    { "dbDate", makeDate, 1} ,
    { "dbiDate", makeDate, 1} ,
    { "dbRaw", makeRaw, 1} ,
    { "dbiRaw", makeRaw, 1} ,
    {0,     0}        /* Sentinel */
};

/* dbi object definition */

typedef struct {
	PyObject_HEAD
	PyObject	*dbi_attr;	/* Attributes dictionary */
} dbiobject;

staticforward PyTypeObject Dbitype;

#define is_dbiobject(v)		((v)->ob_type == &Dbitype)

static dbiobject *
newdbiobject()
{
	dbiobject *xp;
	xp = PyObject_NEW(dbiobject, &Dbitype);
	if (xp == NULL)
		return NULL;
	xp->dbi_attr = NULL;
	return xp;
}

/* Xx methods */

static void
dbi_dealloc(xp)
	dbiobject *xp;
{
	Py_XDECREF(xp->dbi_attr);
	PyMem_DEL(xp);
}

static PyObject *
dbi_getattr(xp, name)
	dbiobject *xp;
	char *name;
{
    if(!strcmp(name, "__dict__"))
    {
        PyObject *k, *v, *d;
        int i;
        PyMethodDef *m;
        d=PyDict_New();
        if(xp->dbi_attr)
        {
            i=0;
            while(PyDict_Next(xp->dbi_attr, &i, &k, &v))
            {
                PyDict_SetItem(d, k, v);
            }
        }

        m=dbiMethods;
        while(m->ml_name)
        {
            PyDict_SetItemString(d, m->ml_name, Py_FindMethod(dbiMethods, (PyObject *)xp, m->ml_name));
            m++;
        }
        return d;
    }
    if (xp->dbi_attr != NULL) {
		PyObject *v = PyDict_GetItemString(xp->dbi_attr, name);
		if (v != NULL) {
			Py_INCREF(v);
			return v;
		}
	}
	return Py_FindMethod(dbiMethods, (PyObject *)xp, name);
}

static int
dbi_setattr(xp, name, v)
	dbiobject *xp;
	char *name;
	PyObject *v;
{
	if (xp->dbi_attr == NULL) {
		xp->dbi_attr = PyDict_New();
		if (xp->dbi_attr == NULL)
			return -1;
	}
	if (v == NULL) {
		int rv = PyDict_DelItemString(xp->dbi_attr, name);
		if (rv < 0)
			PyErr_SetString(PyExc_AttributeError,
			        "delete non-existing xx attribute");
		return rv;
	}
	else
		return PyDict_SetItemString(xp->dbi_attr, name, v);
}

static PyTypeObject Dbitype = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"dbi",			/*tp_name*/
	sizeof(dbiobject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)dbi_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)dbi_getattr, /*tp_getattr*/
	(setattrfunc)dbi_setattr, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};



void initdbi(PyObject *module_dict)
{
    PyObject *d = (PyObject*)newdbiobject();
    dbi_setattr(d, "STRING",
                         DbiString = PyString_FromString("STRING"));
    dbi_setattr(d, "RAW",
                         DbiRaw = PyString_FromString("RAW"));
    dbi_setattr(d, "NUMBER",
                         DbiNumber = PyString_FromString("NUMBER"));
    dbi_setattr(d, "DATE",
                         DbiDate = PyString_FromString("DATE"));
    dbi_setattr(d, "ROWID",
                         DbiRowId = PyString_FromString("ROWID"));
    dbi_setattr(d, "TYPES",
                         Py_BuildValue("(OOOOO)",
                                       DbiString,
                                       DbiRaw,
                                       DbiNumber,
                                       DbiDate,
                                       DbiRowId));
    
    /* Establish errors */
    dbi_setattr(d, "noError",
                         DbiNoError = PyString_FromString("dbi.no-error"));
    dbi_setattr(d, "opError",
                         DbiOpError = PyString_FromString("dbi.operation-error"));
    dbi_setattr(d, "progError",
                         DbiProgError = PyString_FromString("dbi.program-error"));
    dbi_setattr(d, "integrityError",
                         DbiIntegrityError =
                         PyString_FromString("dbi.integrity-error"));
    dbi_setattr(d, "dataError",
                         DbiDataError = PyString_FromString("dbi.data-error"));
    dbi_setattr(d, "internalError",
                         DbiInternalError = 
                         PyString_FromString("dbi.internal-error"));
    /* Setting dbi module to master */
    PyDict_SetItemString(module_dict, "dbi", d);
#ifdef VERBOSE
    printf("informix.dbi\n");
#endif
}

int dbiIsDate(const PyObject *o)
{
    return o->ob_type == &DbiDate_Type;
}

int dbiIsRaw(const PyObject *o)
{
    return o->ob_type == &DbiRaw_Type;
}

int dbiIsRowId(const PyObject *o)
{
    return o->ob_type == &DbiRowId_Type;
}
